/* This is 3D GPU modeling code */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cutil.h>
#include <cuda.h>
#include <pthread.h>

#include "params.c"
#include "cudaErrors.cu"

#define hloc(z,x,y) (z)+((x)+(y)*n2)*n1
#define dloc(z,x,y) (z)+((x)+(y)*n2gpu)*n1gpu
#define dloc2(x,y) (x)+(y)*n2gpu

#define min2(v1,v2) (((v1)<(v2))?(v1):(v2)) 
//a buffer for source function 1/4 GB
#define SOURCE_BUFFER 67108864

__constant__ int n1gpu;
__constant__ int n2gpu;
__constant__ int n3gpu;
__constant__ float bc_agpu;
__constant__ float bc_bgpu;
__constant__ int bc_zmax;
__constant__ int bc_xmax;
__constant__ int bc_ymax;
__constant__ int nt_gpu;
__constant__ float dt_gpu;
__constant__ int wfld_j_gpu;
texture <float, 3, cudaReadModeElementType> vel_ref;
__device__ int min5(int v1,int v2,int v3, int v4, int v5){return min2(min2(min2(v1,v2),min2(v3,v4)),v5);}
cudaArray *vel_gpu_tex;
float *p0_gpu, *p1_gpu, *source_gpu, *cpu_space, *vel_gpu, *s_gpu;

#include "wave_kernel3d8o.cu"

__global__ void inject_kernel(int z, int x, int y, int it, float *source_gpu, float *p)
{
	register int i=blockIdx.x*blockDim.x+threadIdx.x;
	if(i>=1) return;
	p[dloc(z,x,y)] += source_gpu[it];
}

__global__ void slice_kernel(float *p, float *s)
{

        int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=yblock*BLOCKY_SIZE;
	int xl=FAT+threadIdx.y; 		// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 	// global xdir

	for(int yy=0; yy<BLOCKY_SIZE; yy++){
	  s[dloc2(xg,y0+yy)]=p[dloc(FAT,xg,y0+yy)];
	}

}


extern "C" void gpu_init_tex(int *nt, float *dt, int *n1, int *n2, int *n3, float *d1, float *d2, float *d3, float *bc_a, float *bc_b, int *wfld_j, int *gpu_id)
{

    cudaSetDevice(*gpu_id);

    *bc_a=-1.0/(*bc_a);
    *bc_b=log(*bc_b);

    /* allocating memory */
    fprintf(stderr,"Allocating memory... %d mb per wfld\n",(*n1)*(*n2)*(*n3)*sizeof(float)/1000000);
    cuda_call(cudaMalloc((void**) &p0_gpu, (*n1)*(*n2)*(*n3)*sizeof(float)));
    cuda_call(cudaMalloc((void**) &p1_gpu, (*n1)*(*n2)*(*n3)*sizeof(float)));
    cuda_call(cudaMalloc((void**) &s_gpu, (*n2)*(*n3)*sizeof(float)));
    cuda_call(cudaMalloc((void**) &source_gpu,  (*nt)*sizeof(float)));
    fprintf(stderr,"Allocated wavefields...\n");

    cudaChannelFormatDesc cf = cudaCreateChannelDesc<float>();
    cudaExtent const extv = {*n1, *n2, *n3};
    cuda_call(cudaMalloc3DArray(&vel_gpu_tex, &cf, extv));

    cuda_call(cudaMemcpyToSymbol(n1gpu, n1, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n2gpu, n2, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n3gpu, n3, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(bc_agpu, bc_a, sizeof(float)));
    cuda_call(cudaMemcpyToSymbol(bc_bgpu, bc_b, sizeof(float)));
    cuda_call(cudaMemcpyToSymbol(nt_gpu, nt, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(dt_gpu, dt, sizeof(float)));

    float coeffs_cpu[COEFFS_SIZE]=get_coeffs(*d1,*d2,*d3);
    cuda_call(cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float)));

    cuda_call(cudaMemcpyToSymbol(wfld_j_gpu,  wfld_j,   sizeof(int)));
}
 

extern "C" void gpu_wave_prop_tex(int *sou_z, int *sou_x, int *sou_y, float *source_cpu, float *cpu_space, float *data)
{
    int nt, n1, n2, n3, wfld_j;
    float *ptemp, dt;

    cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n3, n3gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&dt, dt_gpu, sizeof(float)));



    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*sizeof(float), cudaMemcpyHostToDevice));

    cuda_call(cudaMemset(p0_gpu, 0,n1*n2*n3*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,n1*n2*n3*sizeof(float)));
    cuda_call(cudaMemset(s_gpu,  0, n2*n3*sizeof(float)));

    cudaExtent const extv = {n1, n2, n3};

    cudaMemcpy3DParms cp = {0};
    cp.srcPtr = make_cudaPitchedPtr((void*)cpu_space, n1*sizeof(float), n2, n3);
    cp.dstArray = vel_gpu_tex;
    cp.extent = extv;
    cp.kind = cudaMemcpyHostToDevice;
    cuda_call( cudaMemcpy3D(&cp));

    vel_ref.filterMode = cudaFilterModeLinear;
    vel_ref.addressMode[0] = cudaAddressModeClamp;
    vel_ref.addressMode[1] = cudaAddressModeClamp;
    vel_ref.addressMode[2] = cudaAddressModeClamp;

    cuda_call(cudaBindTextureToArray(vel_ref, vel_gpu_tex));

    cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));


    int nblocksz=(n1-2*FAT)/BLOCKZ_SIZE;
    int nblocksx=(n2-2*FAT)/BLOCKX_SIZE;
    int nblocksy=(n3-2*FAT)/BLOCKY_SIZE;
    int zmax=nblocksz*BLOCKZ_SIZE+2*FAT;
    int xmax=nblocksx*BLOCKX_SIZE+2*FAT;
    int ymax=nblocksy*BLOCKY_SIZE+2*FAT;
    cuda_call(cudaMemcpyToSymbol(bc_zmax, &zmax, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(bc_xmax, &ymax, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(bc_ymax, &xmax, sizeof(int)));

    dim3 dimGrid(nblocksz*nblocksy,nblocksx);
    dim3 dimBlock(BLOCKZ_SIZE, BLOCKX_SIZE);

    dim3 dimGrid_out(nblocksx,nblocksy);


    fprintf(stderr, "Dimensions %d %d %d\n", n1, FAT, BLOCKZ_SIZE);
    fprintf(stderr, "Number of blocks :[%d, %d] * %d\n", nblocksz, nblocksx, nblocksy);
    fprintf(stderr,"  sou position for injection:  %d, %d, %d  \n",*sou_z,*sou_x,*sou_y);


    for (int it=0; it<nt; it++)
    {
        kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, *sou_x, *sou_y, it, source_gpu, p1_gpu));
        kernel_exec(wave_kernel_tex<<<dimGrid, dimBlock>>>(nblocksz,p0_gpu, p1_gpu, p0_gpu));
        kernel_exec(damp_kernel<<<dimGrid, dimBlock>>>(nblocksz,p0_gpu,p1_gpu));

        if(it%wfld_j==0)
        {
	    kernel_exec(slice_kernel<<<dimGrid, dimBlock>>>(p1_gpu,s_gpu));
            cuda_call(cudaMemcpy(&data[(it/wfld_j)*n2*n3], s_gpu, n2*n3*sizeof(float), cudaMemcpyDeviceToHost));
        }

        ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;

    }
//  cuda_call( cudaFree(p0_gpu));
//  cuda_call( cudaFree(p1_gpu));
// cuda_call( cudaFree(s_gpu) );
//  cuda_call( cudaFree(source_gpu) );
fprintf(stderr,"Finished gpu shot propagation \n");
//  cuda_call( cudaFree(ptemp) );
//  cuda_call( cudaFree(vel_gpu_tex));
//  cuda_call( cudaFree(vel_ref));

}
